package org.jastadd.relast.compiler;

import beaver.Parser;
import org.jastadd.relast.ast.GrammarFile;
import org.jastadd.relast.ast.Program;
import org.jastadd.relast.compiler.options.CommandLine;
import org.jastadd.relast.compiler.options.CommandLine.CommandLineException;
import org.jastadd.relast.compiler.options.Option;
import org.jastadd.relast.compiler.options.StringOption;
import org.jastadd.relast.parser.RelAstParser;
import org.jastadd.relast.scanner.RelAstScanner;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.nio.file.DirectoryStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;

public class Compiler {

  private StringOption optionOutputDir;
  private StringOption optionInputDir;

  private ArrayList<Option<?>> options;
  private CommandLine commandLine;

  public Compiler() {
    options = new ArrayList<>();
    addOptions();
  }

  public static void main(String[] args) {
    try {
      new Compiler().run(args);
    } catch (CommandLineException | CompilerException e) {
      System.err.println(e.getMessage());
      System.exit(-1);
    }
  }

  public int run(String[] args) throws CommandLineException, CompilerException {
    options = new ArrayList<>();
    addOptions();
    commandLine = new CommandLine(options);
    commandLine.parse(args);

    Path inputPath;
    if (optionInputDir.isSet()) {
      inputPath = Paths.get(optionInputDir.getValue());
    } else {
      inputPath = Paths.get(".");
      printMessage("No input dir is set. Assuming current directory '" + inputPath.toAbsolutePath().toString() + "'.");
    }

    if (!inputPath.toFile().exists()) {
      printMessage("Input path '" + inputPath.toAbsolutePath().toString() + "' does not exist. Exiting...");
      System.exit(-1);
    } else if (!inputPath.toFile().isDirectory()) {
      printMessage("Input path '" + inputPath.toAbsolutePath().toString() + "' is not a directory. Exiting...");
      System.exit(-1);
    }

    Path outputPath; // should not be used, but otherwise there is a compiler warning
    if (optionOutputDir.isSet()) {
      outputPath = Paths.get(optionOutputDir.getValue());
    } else {
      outputPath = Paths.get("./gen/");
      printMessage("No output dir is set. Assuming '" + outputPath.toAbsolutePath().toString() + "'.");
    }

    if (outputPath.toFile().exists() && !outputPath.toFile().isDirectory()) {
      printMessage("Output path '" + inputPath.toAbsolutePath().toString() + "' exists, but is not a directory. Exiting...");
    }

    printMessage("Running RelAST Preprocessor");

    List<String> otherArgs = commandLine.getArguments();
    if (!otherArgs.isEmpty()) {
      printMessage("Unsupported arguments will be ignored: " + otherArgs);
    }

    Program program = parseProgram(inputPath);

    printMessage("Writing output files");

    for (GrammarFile grammarFile : program.getGrammarFileList()) {
      // TODO decide and document what the file name should be, the full path or a simple name?
      writeToFile(outputPath + grammarFile.getFileName() + "/Grammar.relast", grammarFile.generateAbstractGrammar());
    }

    writeToFile(outputPath + "/Grammar.relast", program.generateAbstractGrammar());
    return 0;
  }

  private void printMessage(String message) {
    System.out.println(message);
  }

  private void writeToFile(String filename, String str) throws CompilerException {
    try {
      PrintWriter writer = new PrintWriter(filename);
      writer.print(str);
      writer.close();
    } catch (Exception e) {
      throw new CompilerException("Could not write to file " + filename, e);
    }
  }

  private void addOptions() {
    optionOutputDir = addOption(new StringOption("outputDir", "target directory for the generated files."));
    optionInputDir = addOption(new StringOption("inputDir", "input directory."));
  }

  private <OptionType extends Option<?>> OptionType addOption(OptionType option) {
    options.add(option);
    return option;
  }

  private Program parseProgram(Path inputPath) throws CompilerException {
    Program program = new Program();

    try (DirectoryStream<Path> stream = Files.newDirectoryStream(inputPath, "*.relast")) {
      RelAstParser parser = new RelAstParser();
      stream.forEach(path -> {
        try (BufferedReader reader = Files.newBufferedReader(path)) {
          RelAstScanner scanner = new RelAstScanner(reader);
          GrammarFile inputGrammar = (GrammarFile) parser.parse(scanner);
          inputGrammar.dumpTree(System.out);
          program.addGrammarFile(inputGrammar);
          inputGrammar.treeResolveAll();
        } catch (IOException | Parser.Exception e) {
          printMessage("Could not parse grammar file " + path);
          e.printStackTrace();
        }
      });
    } catch (IOException e) {
      printMessage("Unable to iterate over input path '" + inputPath.toAbsolutePath().toString() + "'. Exiting...");
      e.printStackTrace();
      System.exit(-1);
    }


    return program;
  }

  protected int error(String message) {
    System.err.println("Error: " + message);
    System.err.println();
    System.err.println("Usage: java -jar relast.jar [--option1] [--option2=value] ...  <filename1> <filename2> ... ");
    System.err.println("Options:");
    System.err.print(commandLine.printOptionHelp());
    return 1;
  }

  public static class CompilerException extends Exception {
    CompilerException(String message, Throwable cause) {
      super(message, cause);
    }
  }
}

